#!/usr/bin/python
# -*- coding: utf-8 -*-
# $Id$
#
"""This is a simulation of sharks and fish wageing an ecological war
on the toroidal planet Wa-Tor. It comes from an article published by
Dewdney in December 1984 issue of Scientific American."""
__details__ = """
Environment:
1. The world is a toroid that is displayed as a rectangular grid.
    a. The top and bottom edges are joined to each other: moving off the top
       edge puts a shark or fish on the bottom edge.
    b. The left and right edges are joined to each other: moving off the right
       edge puts a shark or fish on the left edge.
2. An initial number of shark and fish are randomly placed in the world.
3. Sharks and fish have a gestation period which is the number of turns before
   reproduction.
4. Sharks have a starvation period which is the number of turns must eat
   before the shark dies.

Behavior Rules (each round):
1. Fish
    a. A fish moves in a random direction: north, south, east, or west if the
       location is clear.
    b. If a fish moves and gestation period has expired then a new fish is
       left in the previous location.
    c. If a fish reproduced then the gestation clock is reset.
2. Shark
    a. A shark randomly chooses a location to move first from locations that
       have fish, and if no fish from empty locations.
    b. If the location has a fish then the shark eats the fish and the hunger
       clock is reset.
    c. If a shark moves and gestation period has expired then a new shark is
       left in the previous location.
    d. If a shark reproduced then the gestation clock is reset.
    e. If a shark's hunger clock expires then the shark dies."""
__references__ = """
References:
    Dewdney, A. K. (1984). Sharks and fish wage an ecological war on the
    toroidal planet Wa-Tor. Scientific American, December 1984, 14--20.

    @Article{Dewdney:1984:CRS,
      author =       "A. K. Dewdney",
      title =        "Computer Recreations: {Sharks} and fish wage an
             ecological war on the toroidal planet Wa-Tor",
      journal =      "Scientific American",
      volume =       "251",
      number =       "6",
      pages =        "14--22",
      month =        dec,
      year =         "1984",
      CODEN =        "SCAMAC",
      ISSN =         "0036-8733",
      bibsource =    "Ai/alife.bib",
      note =         "Description of program for simulating predator-prey
             dynamics.",
      keywords =     "models-computer",
    }
"""
__author__ = ("Lance Finn Helsten",)
__version__ = "1.1.0"
__copyright__ = """Copyright 2006 Lance Finn Helsten (helsten@acm.org)"""
__license__ = """
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import math
import random
import sys


class Wator:
    """Build an instance of the Wator world with a dictionary of options.
    Required Parameters:
        world_size      Number of locations in the world.
        sharks          Number of initial sharks in the world.
        fish            Number of initial fish in the world.
        shark_gestation Shark gestation timer.
        fish_gestation  Fish gestation timer.
        shark_starve    Shark starvation timer.
    """

    INIT_KEYS = [
        "world_size",
        "fish",
        "fish_gestation",
        "sharks",
        "shark_gestation",
        "shark_starve",
    ]

    def __init__(self, **properties):
        # if properties['sharks'] + properties['fish'] > properties['world_size']:
        #    raise ValueError, 'Sharks and fish will not fit in grid.'
        self.empty = Entity(self)
        self.properties = properties
        self.reset()

    def reset(self):
        """Reset the world to the initial round."""
        self.rows = int(math.sqrt(self.properties["world_size"]))
        self.cols = int(self.properties["world_size"] / self.rows)
        self.world = list([None] * self.cols for x in range(0, self.rows))
        self.fish = 0
        self.sharks = 0
        for x in range(0, self.properties["sharks"]):
            self.__place(
                Shark(
                    self,
                    self.properties["shark_gestation"],
                    self.properties["shark_starve"],
                )
            )
        for x in range(0, self.properties["fish"]):
            self.__place(Fish(self, self.properties["fish_gestation"]))
        for r in range(0, self.rows):
            for c in range(0, self.cols):
                if self.world[r][c] is None:
                    self.world[r][c] = self.empty

    def next_round(self):
        """Move to the next round in the simulation."""
        for r in range(0, self.rows):
            for c in range(0, self.cols):
                self.world[r][c].moved = False
        for r in range(0, self.rows):
            for c in range(0, self.cols):
                e = self.world[r][c]
                if e != self.empty and not e.moved:
                    e.move(r, c)

    def __str__(self):
        ret = []
        for r in self.world:
            row = "|" + "".join([str(c) for c in r]) + "|"
            ret.append(row)
        return "\n".join(ret)

    def __repr__(self):
        return f"<Wator instance 0x{hash(self):x}>"

    def __place(self, ent):
        while True:
            r = random.randint(0, self.rows - 1)
            c = random.randint(0, self.cols - 1)
            if self.world[r][c] is None:
                self.world[r][c] = ent
                break


class Entity:
    """Defines an entity that lives in Wator.

    Required Parameters:
        wator       World where the entity lives.
        breed       Number of rounds before the entity breeds: -1 never breeds.
        starve      Number of rounds before the entity starves: -1 never starves.
    """

    def __init__(self, wator, breed=-1, starve=-1):
        self.wator = wator
        self.breed_time = breed
        self.breed_clock = breed
        self.starve_time = starve
        self.starve_clock = starve
        self.moved = False

    def eatable(self):
        """Can this entity be eaten."""
        return False

    def move(self, r, c):
        """Determine how the entity should move and do the move.
        Parameters:
            r The row number of this entity.
            c The column number of this entity.
        """
        # pylint: disable=assignment-from-none
        loc = self.move0(
            (
                self.__get_cell(r - 1, c),
                self.__get_cell(r + 1, c),
                self.__get_cell(r, c - 1),
                self.__get_cell(r, c + 1),
            )
        )
        if loc is not None:
            r0, c0, _ = loc
            self.wator.world[r0][c0] = self.wator.world[r][c]
            self.wator.world[r][c] = self.wator.empty
            self.moved = True
            self.breed(r, c)
            self.starve(r0, c0)
        else:
            self.starve(r, c)

    def __get_cell(self, r, c):
        if r < 0:
            r = self.wator.rows - 1
        if r >= self.wator.rows:
            r = 0
        if c < 0:
            c = self.wator.cols - 1
        if c >= self.wator.cols:
            c = 0
        e = self.wator.world[r][c]
        return (r, c, e)

    def move0(self, dirs):
        """Let sub-class choose which direction to move. The parameters up,
        down, left, and right is a tuple of the row number, col number, and
        entity in the location.
        """
        # pylint: disable=unused-argument
        return None

    def breed(self, r, c):
        """Determine if the entity should breed.

        Parameters:
            r The row number to place a new entity.
            c The column number to place a new entity.
        """
        if self.breed_clock > 0:
            self.breed_clock = self.breed_clock - 1
        if self.breed_clock == 0:
            e = self.breed0()
            self.wator.world[r][c] = e
            self.breed_clock = self.breed_time

    def breed0(self):
        """If entity breeds then inform the subclass."""
        return self.wator.empty

    def starve(self, r, c):
        """Determine if the entity should starve.

        Parameters:
            r The row number to place a new entity.
            c The column number to place a new entity.
        """
        if self.starve_clock > 0:
            self.starve_clock = self.starve_clock - 1
        if self.starve_clock == 0:
            self.starve0(r, c)

    def starve0(self, r, c):
        """If entity starves then inform the subclass."""

    def __str__(self):
        return " "


class Shark(Entity):
    """Defines a Wator shark and its behavior.
    Required Parameters:
        wator       World where the shark lives.
    """

    def __init__(self, wator, gestate, starve):
        Entity.__init__(self, wator, gestate, starve)
        wator.sharks = wator.sharks + 1

    def move0(self, dirs):
        ret = None
        fish = [x for x in dirs if x[2].eatable()]
        if len(fish) > 0:
            ret = random.choice(fish)
        if ret is not None:
            self.wator.fish = self.wator.fish - 1
            self.starve_clock = self.starve_time
        else:
            empty = [x for x in dirs if x[2] == self.wator.empty]
            if len(empty) > 0:
                ret = random.choice(empty)
        return ret

    def breed0(self):
        return Shark(self.wator, self.breed_time, self.starve_time)

    def starve0(self, r, c):
        self.wator.world[r][c] = self.wator.empty
        self.wator.sharks = self.wator.sharks - 1

    def __str__(self):
        return "S"


class Fish(Entity):
    """Defines a Wator fish and its behavior.
    Required Parameters:
        wator       World where the fish lives.
    """

    def __init__(self, wator, gestate):
        Entity.__init__(self, wator, gestate)
        wator.fish = wator.fish + 1

    def eatable(self):
        return True

    def move0(self, dirs):
        ret = None
        empty = [x for x in dirs if x[2] == self.wator.empty]
        if len(empty) > 0:
            random.shuffle(empty)
            ret = empty[0]
        return ret

    def breed0(self):
        return Fish(self.wator, self.breed_time)

    def __str__(self):
        return "F"


def _test():
    pass


def main(args):
    """Main entry point for the simulation."""
    if args.details:
        print(__doc__)
        print(__details__)
        print(__references__)
    elif args.copyright:
        print(__copyright__)
        print(__license__)
    elif args.test:
        _test()
    else:
        try:
            wator = Wator(
                world_size=args.world_size,
                fish=args.fish,
                fish_gestation=args.fish_gestation,
                sharks=args.sharks,
                shark_gestation=args.shark_gestation,
                shark_starve=args.shark_starve,
            )
            print(wator)
            print()
            while wator.fish > 0 and wator.sharks > 0:
                wator.next_round()
                print(wator)
                print()
        except KeyboardInterrupt:
            print("User interruption", file=sys.stderr)
            sys.exit(3)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "--world-size",
        type=int,
        action="store",
        default=1008,
        help="Define the number of grid locations in the world.",
    )
    parser.add_argument(
        "--sharks",
        type=int,
        action="store",
        default=75,
        help="Define the initial number of sharks in the world.",
    )
    parser.add_argument(
        "--fish",
        type=int,
        action="store",
        default=800,
        help="Define the initial number of fish in the world.",
    )
    parser.add_argument(
        "--shark-gestation",
        type=int,
        action="store",
        default=10,
        help="Define the sharks gestation period.",
    )
    parser.add_argument(
        "--fish-gestation",
        type=int,
        action="store",
        default=3,
        help="Define the fish gestation period.",
    )
    parser.add_argument(
        "--shark-starve",
        type=int,
        action="store",
        default=3,
        help="Define the sharks starvation period.",
    )
    parser.add_argument(
        "--details", action="store_true", help="Display program details."
    )
    parser.add_argument(
        "--test", action="store_true", help="Run the internal unit testing."
    )
    parser.add_argument(
        "--copyright",
        action="store_true",
        help="Display copyright and licence information.",
    )
    main(parser.parse_args())
